import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;

public class Demo 
{
    protected DemoEventListener theEventListener = null;
    protected volatile boolean isRunning = false;

    static final int kSleepMax = 1000;
    static final int kSleepMin = 0;
    static final int kSleepInit = 400;

    protected JSlider sliderDemoSpeed;
    protected ButtonGroup buttonView;
    protected JButton buttonStart;
    protected ButtonGroup buttonControl;
    protected JLabel scoreLabel;
    protected JLabel episodeLabel;

    protected Agent theAgent;

    public Demo()
    {
	final int kWorldWidth = 8;
	final int kWorldHeight = 8;

	/* Create the world */
	GridWorld theWorld = new GridWorld(kWorldWidth,kWorldHeight);

	/* Create the agent */
	//theAgent = new SarsaAgent(GridWorld.kNumActions, kWorldWidth*kWorldWidth*kWorldHeight*kWorldHeight);
	theAgent = new QlearningAgent(GridWorld.kNumActions, kWorldWidth*kWorldWidth*kWorldHeight*kWorldHeight);

	System.out.println("#The agent is a " + theAgent.getClass().getName());

	/* Setup the world states */
	//theWorld.setStartPosition(new GridPoint(0,0));
	theWorld.setGoalPosition(new GridPoint(3,3));
	//theWorld.setEnemyPosition(new GridPoint(7,5));
	//theWorld.setAgentPosition(new GridPoint(0,0));
	
	theWorld.reset();
	theWorld.setAgent(theAgent);

	/* Create the renderer */
	GridRenderer2D theRenderer2D = new GridRenderer2D();
	GridRenderer3D theRenderer3D = new GridRenderer3D();

	/* Tell the event listener about the world and the renderer */	
	this.theEventListener = new DemoEventListener();
	this.theEventListener.setWorld(theWorld);
	this.theEventListener.setAgent(theAgent);
	this.theEventListener.setRenderer2D(theRenderer2D);
	this.theEventListener.setRenderer3D(theRenderer3D);
	this.theEventListener.setRenderMode(DemoEventListener.kRenderMode2D);

	this.isRunning = true;
    }

    public JMenuBar createMenuBar() 
    {
	JMenuBar mb = new JMenuBar();
	JMenu mazeMenu = new JMenu("Maze");
	//JMenu scoreMenu = new JMenu("Score");
	JMenu aiMenu = new JMenu("AI Settings");
	
	JMenuItem mazeMenuErase = new JMenuItem("Reset walls");
	JMenuItem mazeMenuLoad = new JMenuItem("Load default walls");

	JMenuItem scoreMenuView = new JMenuItem("View history");
	JMenuItem scoreMenuGraph = new JMenuItem("Graph performance");

	//I really should do this in a loop, oh well
	JMenu aiMenuAlpha = new JMenu("Change learning rate");
	JRadioButtonMenuItem alpha90 = new JRadioButtonMenuItem(".9 (practically ignore the past)");
	ButtonGroup alphas = new ButtonGroup();
	aiMenuAlpha.add(alpha90);
	alphas.add(alpha90);
	alpha90.setActionCommand("alpha90");
	alpha90.addActionListener(theEventListener);
	JRadioButtonMenuItem alpha50 = new JRadioButtonMenuItem(".5");
	aiMenuAlpha.add(alpha50);
	alphas.add(alpha50);
	if(theAgent.getAlpha()==.5)
	    alpha50.setSelected(true);
	alpha50.setActionCommand("alpha50");
	alpha50.addActionListener(theEventListener);
	JRadioButtonMenuItem alpha10 = new JRadioButtonMenuItem(".1 (standard)");
	aiMenuAlpha.add(alpha10);
	alphas.add(alpha10);
	if(theAgent.getAlpha()==.1)
	    alpha10.setSelected(true);
	alpha10.setActionCommand("alpha10");
	alpha10.addActionListener(theEventListener);
	JRadioButtonMenuItem alpha1 = new JRadioButtonMenuItem(".01 (learn slowly but surely)");
	aiMenuAlpha.add(alpha1);
	alphas.add(alpha1);
	if(theAgent.getAlpha()==.01)
	    alpha10.setSelected(true);
	alpha1.setActionCommand("alpha1");
	alpha1.addActionListener(theEventListener);
	aiMenu.add(aiMenuAlpha);

	JMenu aiMenuEpsilon = new JMenu("Change exploration rate");
	JRadioButtonMenuItem epsilon90 = new JRadioButtonMenuItem(".9 (hardly ever do the right thing)");
	ButtonGroup epsilons = new ButtonGroup();
	aiMenuEpsilon.add(epsilon90);
	epsilons.add(epsilon90);
	if(theAgent.getEpsilon()==0.9)
	    epsilon90.setSelected(true);
	epsilon90.setActionCommand("epsilon90");
	epsilon90.addActionListener(theEventListener);
	JRadioButtonMenuItem epsilon50 = new JRadioButtonMenuItem(".5");
	aiMenuEpsilon.add(epsilon50);
	epsilons.add(epsilon50);
	if(theAgent.getEpsilon()==0.5)
	    epsilon50.setSelected(true);
	epsilon50.setActionCommand("epsilon50");
	epsilon50.addActionListener(theEventListener);
	JRadioButtonMenuItem epsilon10 = new JRadioButtonMenuItem(".1 (explore every ten moves)");
	aiMenuEpsilon.add(epsilon10);
	epsilons.add(epsilon10);
	if(theAgent.getEpsilon()==0.1)
	    epsilon10.setSelected(true);
	epsilon10.setActionCommand("epsilon10");
	epsilon10.addActionListener(theEventListener);
	JRadioButtonMenuItem epsilon1 = new JRadioButtonMenuItem(".01 (hardly ever explore)");
	aiMenuEpsilon.add(epsilon1);
	epsilons.add(epsilon1);
	if(theAgent.getEpsilon()==0.01)
	    epsilon1.setSelected(true);
	epsilon1.setActionCommand("epsilon1");
	epsilon1.addActionListener(theEventListener);
	aiMenu.add(aiMenuEpsilon);

	JMenuItem aiMenuValue = new JMenuItem("Reset memory");
	
	aiMenuEpsilon.addActionListener(theEventListener);
	aiMenu.add(aiMenuValue);
	aiMenuValue.addActionListener(theEventListener);


	mb.add(aiMenu);
	return mb;
    }

    public JPanel createButtonPanel()
    {
	/*
	JPanel panel = new JPanel();

	JButton button2D = new JButton("2D");
	JButton button3D = new JButton("3D");
	JButton buttonOverlay = new JButton("Overlay");
	JButton buttonStart = new JButton("Start/Stop");
	JButton buttonStep = new JButton("Step");
	JButton buttonFaster = new JButton("Faster");
	JButton buttonSlower = new JButton("Slower");
	JButton buttonPlay = new JButton("AI/Play");

	button2D.addActionListener(theEventListener);
	button3D.addActionListener(theEventListener);
	buttonOverlay.addActionListener(theEventListener);
	buttonStart.addActionListener(theEventListener);
	buttonStep.addActionListener(theEventListener);
	buttonFaster.addActionListener(theEventListener);
	buttonSlower.addActionListener(theEventListener);
	buttonPlay.addActionListener(theEventListener);

	panel.add(button2D);
	panel.add(button3D);
	panel.add(buttonOverlay);
	panel.add(buttonStart);
	panel.add(buttonStep);
	panel.add(buttonFaster);
	panel.add(buttonSlower);
	panel.add(buttonPlay);
	
	return panel;
	*/

        JPanel panel = new JPanel(new GridLayout(1,0, 30, 5));
	JPanel viewPanel = new JPanel(new GridLayout(3,0));
	JPanel speedPanel = new JPanel(new GridLayout(2,0));
	JPanel startPanel = new JPanel(new GridLayout(2,0, 0, 5));
	JPanel scorePanel = new JPanel(new GridLayout(2,0));

	this.buttonStart = new JButton("Start");
        JButton buttonStep = new JButton("Step");
	startPanel.add(buttonStart);
	startPanel.add(buttonStep);

        this.buttonControl = new ButtonGroup();
        JRadioButton button2D = new JRadioButton("Player 2D");
	button2D.setActionCommand("2D");
        JRadioButton button3D = new JRadioButton("Player 3D");
	button3D.setActionCommand("3D");
        JRadioButton buttonAI = new JRadioButton("AI");
        button2D.setSelected(true);
        this.buttonControl.add(button2D);
	this.buttonControl.add(button3D);
        this.buttonControl.add(buttonAI);
	viewPanel.add(button2D);
	viewPanel.add(button3D);
	viewPanel.add(buttonAI);
	

	this.episodeLabel = new JLabel("Try 0");
	this.scoreLabel = new JLabel("Score 0");
	scorePanel.add(episodeLabel);
	scorePanel.add(scoreLabel);
      
	speedPanel.add(new JLabel("  AI Delay:"));
        this.sliderDemoSpeed = new JSlider(JSlider.HORIZONTAL, kSleepMin, kSleepMax, kSleepInit);
	sliderDemoSpeed.setMajorTickSpacing((kSleepMax-kSleepMin)/2);
	sliderDemoSpeed.setMinorTickSpacing((kSleepMax-kSleepMin)/4);
	sliderDemoSpeed.setPaintTicks(true);
	sliderDemoSpeed.setPaintLabels(true);
	speedPanel.add(sliderDemoSpeed);
        
	this.theEventListener.setScoreLabel(this.scoreLabel);
	this.theEventListener.setEpisodeLabel(this.episodeLabel);
	this.theEventListener.setStartButton(this.buttonStart);

        this.sliderDemoSpeed.addChangeListener(theEventListener);
        button2D.addActionListener(theEventListener);
        button3D.addActionListener(theEventListener);
        buttonAI.addActionListener(theEventListener);
	buttonStart.addActionListener(theEventListener);
        buttonStep.addActionListener(theEventListener);

	panel.add(viewPanel);
	panel.add(speedPanel);
        panel.add(startPanel);
        panel.add(scorePanel);

        return panel;
    }

    public JPanel createControlPanel()
    {
	return this.createButtonPanel();
    }

    public GLCanvas createOpenGLCanvas()
    {
	GLCanvas canvas = new GLCanvas();

	canvas.addGLEventListener(this.theEventListener);
	canvas.addMouseListener(this.theEventListener);
	canvas.addMouseMotionListener(this.theEventListener);
	canvas.addKeyListener(this.theEventListener);
	return canvas;
    }

    public void createAndShowGUI()
    {
	JFrame frame = new JFrame("Demo Framework");

	JPanel panel = this.createControlPanel();
	JMenuBar menuBar = this.createMenuBar();
      	GLCanvas canvas = this.createOpenGLCanvas();

	this.theEventListener.setCanvas(canvas);
	
	Container content = frame.getContentPane();
	content.setLayout(new BorderLayout());

	frame.getContentPane().add(canvas, BorderLayout.CENTER);
	frame.getContentPane().add(panel, BorderLayout.SOUTH);
	frame.setJMenuBar(menuBar);
	frame.pack();
	
	frame.setSize(600, 600);

	frame.addWindowListener(new WindowAdapter()
	{
	    public void windowClosing(WindowEvent e)
	    {
		isRunning = false;
		System.exit(0);
	    }
	});
			
	frame.setVisible(true);
	canvas.requestFocus();

	while( this.isRunning )
	{
	    this.theEventListener.update();
	    canvas.display();
	}
    }

    public void run()
    {
	this.createAndShowGUI();
    }

    public static void main(String[] args)
    {
	Demo theDemo = new Demo();
	theDemo.run();
    }
}
