import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.Point;

import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import javax.swing.JSlider;
import javax.swing.JLabel;
import javax.swing.JButton;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLAutoDrawable;

public class DemoEventListener implements GLEventListener, 
					  MouseListener, 
					  MouseMotionListener,
					  ActionListener,
					  KeyListener,
					  ChangeListener
{
    public static final int kRenderMode2D = 0;
    public static final int kRenderMode3D = 1;
    public static final int kRenderModeOverlay = 2;

    public boolean isDemoRunning;
    public boolean isAgentRunning;

    protected int theRenderMode;

    protected GLCanvas theCanvas;
    protected GridRenderer2D theRenderer2D;
    protected GridRenderer3D theRenderer3D;
    protected GridWorld theWorld;
    protected Agent theAgent;

    protected GridPoint mousePoint;
    protected boolean isDragging;
    protected int theObject;

    protected int theRLState;
    protected Integer theAction;
    protected Integer theObservation;
    protected Double theReward;
    protected Boolean isTerminal;
    protected Object[] theObservationRewardTerminal;

    protected int theSleepTime;
    protected long theStartTime;
    protected long theElapsedTime;

    protected float theAngle;
    protected boolean[] theKeys;

    protected final int kUpdateAgent = 0;
    protected final int kUpdateEnemy = 1;
    protected int doUpdateFor;
    protected boolean isGoalEligible = true;
    protected GridPoint thePreviousPoint;
    protected GridPoint theLookPoint;

    protected JLabel scoreLabel;
    protected JLabel episodeLabel;
    protected JButton startButton;

    public DemoEventListener()
    {
	theRLState = 0;
	theAction = new Integer(0);
	theObservation = new Integer(0);
	theReward = new Double(0.0);
	isTerminal = new Boolean(false);
	theObservationRewardTerminal = new Object[3];
	
	isDemoRunning = false;
	isAgentRunning = false;

	this.theSleepTime = 200;
	this.theStartTime = System.currentTimeMillis();
	this.theKeys = new boolean[256];
    }

    public void setCanvas(GLCanvas canvas)
    {
	this.theCanvas = canvas;
    }

    public void setRenderer2D(GridRenderer2D theRenderer)
    {
	this.theRenderer2D = theRenderer;
	this.theRenderer2D.drawAgentValues(false);
    }

    public void setRenderer3D(GridRenderer3D theRenderer)
    {
	this.theRenderer3D = theRenderer;
    }

    public void setRenderMode(int theRenderMode)
    {
	this.theRenderMode = theRenderMode;
    }

    public void setWorld(GridWorld theWorld)
    {
	this.theWorld = theWorld;
    }

    public void setAgent(Agent theAgent)
    {
	this.theAgent = theAgent;
    }

    public void setScoreLabel(JLabel theLabel)
    {
	this.scoreLabel=theLabel;
    }

    public void setEpisodeLabel(JLabel theLabel)
    {
	this.episodeLabel=theLabel;
    }

    public void setStartButton(JButton theButton)
    {
	this.startButton=theButton;
    }

    /* GLEventListener Interface */
    public void init(GLAutoDrawable glDrawable)
    {
	Renderer.setGL(glDrawable.getGL());
	Renderer.setGLU(new GLU());
    }

    public void reshape(GLAutoDrawable glDrawable, int x, int y, int width, int height)
    {
	float theAspectRatio = (height == 0) ? width : width/(float)height;
	Camera theCamera2D = this.theRenderer2D.getCamera();
	Camera theCamera3D = this.theRenderer3D.getCamera();

	Viewport theViewport2D = new Viewport(0,0,width,height);
	Viewport theViewport3D = new Viewport(0,0,width,height);

	Projection theProjection2D = new Orthographic2D(width,height);
	Projection theProjection3D = new Perspective(65.0f, theAspectRatio, 0.25f, 100.0f);

	theCamera2D.setViewport(theViewport2D);
	theCamera3D.setViewport(theViewport3D);

	theCamera2D.setProjection(theProjection2D);
	theCamera3D.setProjection(theProjection3D);
    }

    public void display2D()
    {
	this.theRenderer2D.getCamera().apply();

	Renderer.GL.glPushMatrix();
	this.theRenderer2D.present(theWorld);
	Renderer.GL.glPopMatrix();
    }

    public void display3D()
    {
	this.theRenderer3D.getCamera().apply();

	Renderer.GL.glPushMatrix();
	this.theRenderer3D.present(theWorld);
	Renderer.GL.glPopMatrix();
    }

    public void displayOverlay()
    {
	Viewport theViewport2D = this.theRenderer2D.getCamera().getViewport();

	int x = theViewport2D.width  * 4 / 5;
	int y = theViewport2D.height * 4 / 5;
	int w = theViewport2D.width  * 1 / 5;
	int h = theViewport2D.height * 1 / 5;

	this.theRenderer3D.GL.glPushMatrix();
	this.theRenderer3D.getCamera().apply();
	this.theRenderer3D.present(this.theWorld);
	this.theRenderer3D.GL.glPopMatrix();

	this.theRenderer2D.GL.glPushMatrix();
	this.theRenderer2D.getCamera().apply();
	this.theRenderer3D.GL.glViewport(x,y,w,h);
	this.theRenderer2D.setAngle(theAngle);
	this.theRenderer2D.present(this.theWorld);
	this.theRenderer2D.setAngle(0.0);
	this.theRenderer2D.GL.glPopMatrix();
    }

    public void display(GLAutoDrawable glDrawable)
    {
	if (this.theRenderMode == kRenderMode2D)
	    this.display2D();
	
	else if (this.theRenderMode == kRenderMode3D)
	    this.display3D();
	
	else if (this.theRenderMode == kRenderModeOverlay)
	    this.displayOverlay();
    }

    public void displayChanged(GLAutoDrawable glDrawable, boolean modeChanged, boolean deviceChanged)
    {
    }

    /* ChangeListener Interface */
    public void stateChanged(ChangeEvent e)
    {
        JSlider source = (JSlider)e.getSource();
        if(!source.getValueIsAdjusting())
	{
	    int pt = (int)source.getValue();
	    this.theSleepTime = pt;
	}
	this.theCanvas.requestFocus();
    }

    /* ActionListener Interface */
    public void actionPerformed(ActionEvent e)
    {
	if (e.getActionCommand().equals("2D")) {
	    this.setRenderMode(kRenderMode2D);
	    this.isAgentRunning = false;
	    this.theRenderer2D.drawAgentValues(false);
	}
	else if (e.getActionCommand().equals("3D")) {
	    this.setRenderMode(kRenderModeOverlay);
	    this.isAgentRunning = false;
	    this.theRenderer2D.drawAgentValues(false);
	}
	else if (e.getActionCommand().equals("Start")) {
            this.isDemoRunning = true;
	    startButton.setText("Stop");
	}
        else if (e.getActionCommand().equals("Stop")) {
            this.isDemoRunning = false;
	    startButton.setText("Start");
	}
        else if (e.getActionCommand().equals("AI"))
	{
	    this.theRenderer2D.drawAgentValues(true);
            this.isAgentRunning = true;
	}

        else if (e.getActionCommand().equals("Player"))
	{
	    this.theRenderer2D.drawAgentValues(false);
            this.isAgentRunning = false;
	}
	else if (e.getActionCommand().equals("Slower"))
	{
	    this.theSleepTime += 100;
	    if (this.theSleepTime > 3000)
		this.theSleepTime = 3000;
	}
	
	else if (e.getActionCommand().equals("Faster"))
	{
	    this.theSleepTime -= 100;
	    if (this.theSleepTime < 0)
		this.theSleepTime = 0;
	}

	else if (e.getActionCommand().equals("Step"))
	{
	    this.isDemoRunning = true;
	    this.update();
	    this.isDemoRunning = false;
	}
	
	else if (e.getActionCommand().equals("Start/Stop"))
	{
	    this.isDemoRunning = !this.isDemoRunning;
	}

	else if (e.getActionCommand().equals("AI/Play"))
	{
	    this.isAgentRunning = !this.isAgentRunning;
	}

	else if (e.getActionCommand().equals("alpha1")) {
	    theAgent.setAlpha(.01);
	}
	else if (e.getActionCommand().equals("alpha10")) {
	    theAgent.setAlpha(.1);
	}
	else if (e.getActionCommand().equals("alpha50")) {
	    theAgent.setAlpha(.5);
	}
	else if (e.getActionCommand().equals("alpha90")) {
	    theAgent.setAlpha(.9);
	}
	else if (e.getActionCommand().equals("epsilon1")) {
	    theAgent.setEpsilon(.01);
	}
	else if (e.getActionCommand().equals("epsilon10")) {
	    theAgent.setEpsilon(.1);
	}
	else if (e.getActionCommand().equals("epsilon50")) {
	    theAgent.setEpsilon(.5);
	}
	else if (e.getActionCommand().equals("epsilon90")) {
	    theAgent.setEpsilon(.9);
	}
	else if (e.getActionCommand().equals("Reset memory")) {
	    theAgent.resetValue();
	}
	this.theCanvas.requestFocus();
    }

    public void keyTyped(KeyEvent e)
    {
    }

    public void keyPressed(KeyEvent e)
    {
	this.theKeys[e.getKeyCode()] = true;
    }

    public void keyReleased(KeyEvent e)
    {
	this.theKeys[e.getKeyCode()] = false;
    }

    /* MouseListener Interface */
    public void mousePressed(MouseEvent e) 
    {
	if (this.theRenderMode == kRenderMode2D)
	{
	    this.isDragging = false;
	    this.mousePoint = this.getLocation(e.getPoint()); 
	    this.theObject = this.getObjectAtPoint(this.mousePoint);
	}
    }

    public void mouseReleased(MouseEvent e) 
    {
	if (this.theRenderMode == kRenderMode2D)
	{
	    GridPoint thePoint = this.getLocation(e.getPoint());
	    if (thePoint.equals(mousePoint) && this.isDragging == false)
	    {
		if (this.theWorld.isWall(thePoint))
		    this.theWorld.removeWall(thePoint);
		else 
		    this.theWorld.addWall(thePoint);
		
	    }
	    this.isDragging = false;
	}
    }

    public void mouseEntered(MouseEvent e) 
    {
    }

    public void mouseExited(MouseEvent e) 
    {
    }

    public void mouseClicked(MouseEvent e) 
    {
    }

    public void mouseMoved(MouseEvent e)
    {
    }

    public void mouseDragged(MouseEvent e)
    {
	if (this.theRenderMode == kRenderMode2D)
	{
	    GridPoint thePoint = this.getLocation(e.getPoint());
	    
	    this.isDragging = true;
	    
	    if (this.theObject == GridWorld.kAgent)
		theWorld.setAgentPosition(thePoint);
	    
	    else if (this.theObject == GridWorld.kEnemy)
		theWorld.setEnemyPosition(thePoint);
	    
	    else if (this.theObject == GridWorld.kGoal)
		theWorld.setGoalPosition(thePoint);
	    
	    else if (this.theObject == GridWorld.kEmpty)
		theWorld.addWall(thePoint);
	    
	    else if (this.theObject == GridWorld.kWall)
		theWorld.removeWall(thePoint);
	    
	    mousePoint = thePoint;
	}
    }

    public GridPoint getLocation(Point thePoint)
    {
	/* To be in edit mode, we're in a 2D view, so the renderer is a 2D renderer */
	final Viewport theViewport = this.theRenderer2D.getCamera().getViewport();
	final int gridWidth = this.theWorld.getWidth();
	final int gridHeight = this.theWorld.getHeight();
	final float xp = (thePoint.x * this.theWorld.getWidth()) / (float)theViewport.width;
	final float yp = (thePoint.y * this.theWorld.getHeight()) / (float)theViewport.height;

	int x = (int)xp;
	int y = (int)yp;

	x = Math.max(0, x);
	x = Math.min(x, theWorld.getWidth()-1);

	y = Math.max(0, y);
	y = Math.min(y, theWorld.getHeight()-1);

	return new GridPoint(x,y);
    }

    public int getObjectAtPoint(GridPoint thePoint)
    {
	int theObject = GridWorld.kEmpty;

	if (theWorld.isAgentPosition(thePoint))
	    theObject = GridWorld.kAgent;

	else if (theWorld.isEnemyPosition(thePoint))
	    theObject = GridWorld.kEnemy;

	else if (theWorld.isGoalPosition(thePoint))
	    theObject = GridWorld.kGoal;

	else if (theWorld.isWall(thePoint))
	    theObject = GridWorld.kWall;

	return theObject;
    }

    public void updateCamera()
    {
    }

    public void doPlayerUpdate(long theCurrentTime)
    {
	GridPoint theAgentPosition = theWorld.getAgentPosition();
	GridPoint theEnemyPosition = theWorld.getEnemyPosition();

	// Will be ignored by GridWorld, generating a stay action.
	// It is not a "normal" action because it shouldn't be available to agents.
	int newAction = GridWorld.kNumActions;     
	   
	if (this.theKeys[KeyEvent.VK_UP])
	{
	    newAction = GridWorld.kUp;
	    this.theKeys[KeyEvent.VK_UP] = false;
	}
	else if (this.theKeys[KeyEvent.VK_DOWN])
	{
	    newAction = GridWorld.kDown;
	    this.theKeys[KeyEvent.VK_DOWN] = false;
	}
	else if (this.theKeys[KeyEvent.VK_LEFT])
	{
	    newAction = GridWorld.kLeft;
	    this.theKeys[KeyEvent.VK_LEFT] = false;
	}
	else if (this.theKeys[KeyEvent.VK_RIGHT])
	{
	    newAction = GridWorld.kRight;
	    this.theKeys[KeyEvent.VK_RIGHT] = false;
	}
		
	// When the user is in control, calling step doesn't make sense.
	// We have to do the book keeping ourselves.
	
	if (this.theRenderMode == kRenderMode2D)
        {
	    // Apply the new action to the agent
	    this.theWorld.moveAgent(newAction);
	}
	else if (this.theRenderMode == kRenderModeOverlay)
	{	
	    if (newAction == GridWorld.kLeft) theAngle -= Math.PI / 2.0;
	    else if (newAction == GridWorld.kRight) theAngle += Math.PI / 2.0;
    
	    double x = theAgentPosition.x + 0.5;
	    double z = theAgentPosition.y + 0.5;
	    double lx = Math.sin(theAngle);
	    double lz =-Math.cos(theAngle);
	    
	    if (newAction == GridWorld.kUp)
	    {
		x = x + lx;
		z = z + lz;
	    }
	    
	    else if (newAction == GridWorld.kDown)
	    {
		x = x - lx;
		z = z - lz;
	    }

	    if (!isAgentRunning)
	    {
		GridPoint p = new GridPoint((int)x,(int)z);
		if (!theWorld.isWall(new GridPoint(p)))
		{
		    this.theWorld.setAgentPosition(p);
		}
	    }
	    this.theRenderer3D.getCamera().lookAt(x,0,z,x+lx,0,z+lz);
	}
	// Check Terminal Condition
	theAgentPosition = theWorld.getAgentPosition();
	theEnemyPosition = theWorld.getEnemyPosition();

	boolean addedCaptureScore = false;

	if (theAgentPosition.equals(theEnemyPosition))
	{
	    this.isTerminal = Boolean.TRUE;
	    this.theWorld.addScore(-100);
	    addedCaptureScore = true;
	}

	if (theAgentPosition.equals(theWorld.getGoalPosition()))
	{
	    if (this.isGoalEligible)
	    {
		this.theWorld.addScore(5);
		this.isGoalEligible = false;
	    }
	}
	else
	{
	    this.isGoalEligible = true;
	}

	// Move the enemy
	if (this.theElapsedTime > this.theSleepTime)
	{
	    this.theStartTime = theCurrentTime;
	    this.theWorld.moveEnemy();
	}
	
	theAgentPosition = theWorld.getAgentPosition();
	theEnemyPosition = theWorld.getEnemyPosition();
	
	if (theAgentPosition.equals(theEnemyPosition))
	{
	    this.isTerminal = Boolean.TRUE;
	    if (!addedCaptureScore)
		this.theWorld.addScore(-100);
	}
    }

    public void doAgentUpdate(long theCurrentTime)
    {
	final Agent theAgent = theWorld.getAgent();
	if (this.theElapsedTime > this.theSleepTime)
	{
	    if (doUpdateFor == kUpdateAgent)
	    {
		this.theStartTime = theCurrentTime;
		doUpdateFor = kUpdateEnemy;
		
		theAction = (Integer)theAgent.step(theReward,theObservation);
	    }
	    else if (doUpdateFor == kUpdateEnemy)
	    {
		this.theStartTime = theCurrentTime;
		doUpdateFor = kUpdateAgent;
		
		this.theObservationRewardTerminal = this.theWorld.step(theAction);
		this.theObservation = (Integer)theObservationRewardTerminal[0];
		this.theReward      = (Double)theObservationRewardTerminal[1];
		this.isTerminal     = (Boolean)theObservationRewardTerminal[2];
	    }
	}
    }

    public void update()
    {
	final Agent theAgent = theWorld.getAgent();
	final long theCurrentTime = System.currentTimeMillis();

	this.theElapsedTime = theCurrentTime - this.theStartTime;

	if (!isDemoRunning)
	    return;

	if (isDragging)
	    return;

	if (theRLState == 0)
	{
	    if (this.theElapsedTime > this.theSleepTime)
	    {
		this.theStartTime = theCurrentTime;
		this.theWorld.init();
		theAgent.init();
		this.theRLState = 1;
	    }

	    if (this.theRenderMode == kRenderModeOverlay)
		this.updateCamera();
	}
	
	else if (theRLState == 1)
	{
	    if (this.theElapsedTime > this.theSleepTime)
	    {
		// Setup the world for the next episode.
		this.theStartTime = theCurrentTime;
		this.theWorld.reset();
		//System.out.println("Reset!");
		this.isTerminal = Boolean.FALSE;
		this.theObservation = (Integer)this.theWorld.start();
		this.theAction = (Integer)theAgent.start(this.theObservation);
		this.theRLState = 2;
		this.doUpdateFor = kUpdateAgent;
	    }

	    if (this.theRenderMode == kRenderModeOverlay)
		this.updateCamera();
	}

	else if (theRLState == 2)
	{
	    if (isAgentRunning)
	    {
		this.doAgentUpdate(theCurrentTime);
	    }
	    else
	    {
		this.doPlayerUpdate(theCurrentTime);
	    }

	    if (this.isTerminal.equals(Boolean.TRUE))
	    {
		this.theRLState = 3;
	    }
	    
	    if (this.theRenderMode == kRenderModeOverlay)
		this.updateCamera();
	}

	else if (theRLState == 3)
	{
	    if (this.theElapsedTime > this.theSleepTime)
	    {
		this.theStartTime = theCurrentTime;
		if (isAgentRunning)
		    theAgent.end(this.theReward);
		
		this.theRLState = 1;
	    }

	    if (this.theRenderMode == kRenderModeOverlay)
		this.updateCamera();
	}
	this.episodeLabel.setText("Try "+theWorld.getEpisodeNum());
	this.scoreLabel.setText("Score "+theWorld.getScore());
    }
}
