import java.awt.Font;
import java.awt.geom.Rectangle2D;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;
import com.sun.opengl.util.j2d.TextRenderer;

public class GridRenderer2D extends Renderer
{
    protected Font theFont = null;
    protected TextRenderer theTextRenderer = null;
    protected boolean drawValues = false;
    protected double angle = 0.0;

    protected void drawAgentValues(boolean drawValues)
    {
	this.drawValues = drawValues;
    }

    protected void setAngle(double angle)
    {
	this.angle = angle;
    }

    GridRenderer2D()
    {
	this.theFont = new Font("SansSerif", Font.BOLD, 20);
	this.theTextRenderer = new TextRenderer(this.theFont);
    }

    public void present(GridWorld theWorld)
    {
	final Camera camera = this.getCamera();

	// Get the number of pixels per cell (width, height)
	// This way we know how large to draw each cell.
	final float nWidth  = camera.getViewport().width  / (float)theWorld.getWidth();
	final float nHeight = camera.getViewport().height / (float)theWorld.getHeight();
	final float eWidth  = nWidth /20.0f;
	final float eHeight = nHeight/20.0f;

	//camera.apply();

	//GL.glClear(GL.GL_COLOR_BUFFER_BIT);
	GL.glLoadIdentity();
	
	for (int y = 0; y < theWorld.getHeight(); ++y)
	{
	    for (int x = 0; x < theWorld.getWidth(); ++x)
	    {
		final float xOffset = x*nWidth;
		final float yOffset = y*nHeight;

		if (theWorld.isWall(new GridPoint(x,y)))
		    GL.glColor3f(0.3f, 0.44f, 0.55f);
		else if (this.drawValues)
		{
		    double d = theWorld.getValue(new GridPoint(x,y));
		    double g = 1.0 / (1.0 + Math.pow(Math.E, -(1.0 - d)));
		    GL.glColor3d(g, 1.0, g);
		}
		else
		    GL.glColor3d(1.0, 1.0, 1.0);

		this.drawTile(xOffset, yOffset, nWidth, nHeight);
		this.drawPolicy(xOffset, yOffset, nWidth, nHeight);
	    }
	}

	GridPoint agent = theWorld.getAgentPosition();
	GridPoint enemy = theWorld.getEnemyPosition();
	GridPoint start = theWorld.getStartPosition();
	GridPoint goal  = theWorld.getGoalPosition();

	this.drawAgent(agent.x * nWidth, agent.y * nHeight, nWidth, nHeight);
	this.drawEnemy(enemy.x * nWidth, enemy.y * nHeight, nWidth, nHeight);
	this.drawGoal(goal.x * nWidth, goal.y * nHeight, nWidth, nHeight);
    }

    protected void drawTile(float x, float y, float width, float height)
    {
	// Draw the Cell
	GL.glBegin(GL.GL_QUADS);
	GL.glVertex2f(x, y);
	GL.glVertex2f(x + width, y);
	GL.glVertex2f(x + width, y + height);
	GL.glVertex2f(x, y + height);
	GL.glEnd();
    }

    protected void drawLeftWall(float x, float y, float width, float height, float eWidth, float eHeight)
    {
	GL.glColor3f(0.3f, 0.44f, 0.55f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex2f(x + eWidth, y + eHeight);
	GL.glVertex2f(x + eWidth, y + height - eHeight);
	GL.glEnd();
    }

    protected void drawRightWall(float x, float y, float width, float height, float eWidth, float eHeight)
    {
	GL.glColor3f(0.3f, 0.44f, 0.55f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex2f(x + width - eWidth, y + eHeight);
	GL.glVertex2f(x + width - eWidth, y + height - eHeight);
	GL.glEnd();
   }

    protected void drawTopWall(float x, float y, float width, float height, float eWidth, float eHeight)
    {
	GL.glColor3f(0.3f, 0.44f, 0.55f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex2f(x + eWidth, y + height - eHeight);
	GL.glVertex2f(x + width - eWidth, y + height - eHeight);
	GL.glEnd();
    }

    protected void drawBottomWall(float x, float y, float width, float height, float eWidth, float eHeight)
    {
	GL.glColor3f(0.3f, 0.44f, 0.55f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex2f(x + eWidth, y + eHeight);
	GL.glVertex2f(x + width - eWidth, y + eHeight);    
	GL.glEnd();
    }

    public void drawLabel(float x, float y, float width, float height, String theString)
    {
	Rectangle2D theBounds = theTextRenderer.getBounds(theString);
	theTextRenderer.begin3DRendering();
	theTextRenderer.setColor(0.0f, 0.0f, 0.0f, 1.0f);
	theTextRenderer.draw3D(theString, 
			       x + width/2.0f - (float)theBounds.getCenterX(), 
			       y + height/2.0f + (float)theBounds.getCenterY(), 
			       0.0f, 1.0f);
	theTextRenderer.end3DRendering();
    }

    public void drawAgent(float x, float y, float width, float height)
    {
	/* We adhoc determine these values to make the agent slightly smaller than
	   the square it's running in */

	float ew = width  / 4.0f;
	float eh = height / 4.0f;

	double left  = x+ew;
	double right = x+width-ew;
	double top   = y+eh;
	double bottom= y+height-eh;

	double centerx = (left+right)/2.0;
	double centery = (top+bottom)/2.0;
	
	GL.glPushMatrix();
	GL.glLoadIdentity();
	GL.glTranslated(centerx, centery, 0.0);
	GL.glRotated(angle * 180.0 / Math.PI, 0.0, 0.0, 1.0);
	GL.glTranslated(-centerx, -centery, 0.0);

	GL.glColor3f(0.0f, 0.0f, 0.0f);
	GL.glBegin(GL.GL_TRIANGLES);
	GL.glVertex2f(x+ew, y+height-eh);
	GL.glVertex2f(x+width-ew, y+height-eh);
	GL.glVertex2f(x+(width/2.0f), y+eh);
	GL.glEnd();

	GL.glPopMatrix();
    }

    public void drawEnemy(float x, float y, float width, float height)
    {
	float ew = width  / 4.0f;
	float eh = height / 4.0f;

	GL.glColor3f(0.68f, 0.12f, 0.16f);
	GL.glBegin(GL.GL_QUADS);
	GL.glVertex2f(x+ew,       y+eh);
	GL.glVertex2f(x+width-ew, y+eh);
	GL.glVertex2f(x+width-ew, y+height-eh);
	GL.glVertex2f(x+ew,       y+height-eh);
	GL.glEnd();
    }

    public void drawGoal(float x, float y, float width, float height)
    {
	float ew = width  / 3.0f;
	float eh = height / 3.0f;

	GL.glColor3f(1.0f, 1.0f, 0.0f);
	GL.glBegin(GL.GL_QUADS);
	GL.glVertex2f(x+ew,       y+eh);
	GL.glVertex2f(x+width-ew, y+eh);
	GL.glVertex2f(x+width-ew, y+height-eh);
	GL.glVertex2f(x+ew,       y+height-eh);
	GL.glEnd();
    }

    protected void drawPolicy(float x, float y, float width, float height)
    {
	final float xStart = 4.0f/10.0f;
	final float xFinish = 6.0f/10.0f;

	final float yStart = 4.0f/10.0f;
	final float yFinish = 6.0f/10.0f;

	GL.glColor3f(0.5f, 0.5f, 0.5f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex2f(x + (xStart * width), y + (height / 2.0f)); 
	GL.glVertex2f(x + (xFinish * width), y + (height / 2.0f)); 
	GL.glVertex2f(x + (width/2.0f), y + (yStart * height)); 
	GL.glVertex2f(x + (width/2.0f), y + (yFinish * height)); 
	GL.glEnd();
    }
}
