import java.awt.Font;
import java.awt.geom.Rectangle2D;
import java.io.File;
import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureIO;
import com.sun.opengl.util.j2d.TextRenderer;

public class GridRenderer3D extends Renderer
{
    protected Font theFont = null;
    protected TextRenderer theTextRenderer = null;
    protected Texture[] theTextures;
    protected static boolean isLoaded = false;

    GridRenderer3D()
    {
	this.theFont = new Font("SansSerif", Font.BOLD, 20);
	this.theTextRenderer = new TextRenderer(this.theFont, true, true);
    }

    public void loadTextures()
    {
	String path = "textures/";
	String[] filenames = {"stone_wall.jpg", "grass.jpg"};
	this.theTextures = new Texture[filenames.length];

	for (int i = 0; i < filenames.length; ++i)
	{
	    try
	    {
		File theFile = new File(path+filenames[i]);
		theTextures[i] = TextureIO.newTexture(theFile, true);
		theTextures[i].bind();
		theTextures[i].setTexParameteri(GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR_MIPMAP_LINEAR);
		theTextures[i].setTexParameteri(GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR_MIPMAP_LINEAR);
		System.out.println(theFile);
	    }
	    catch(Exception e)
	    {
		System.err.println(e);
		System.exit(0);
	    }
	}
    }

    public void present(GridWorld theWorld)
    {
	final Camera camera = this.getCamera();
	final GridPoint agent = theWorld.getAgentPosition();
	final GridPoint enemy = theWorld.getEnemyPosition();
	final GridPoint start = theWorld.getStartPosition();
	final GridPoint goal  = theWorld.getGoalPosition();

	//camera.apply();

	if (!isLoaded)
	{
	    this.loadTextures();
	    isLoaded = true;
	}
	
	GL.glEnable(GL.GL_TEXTURE_2D);
	GL.glEnable(GL.GL_DEPTH_TEST);
	GL.glClearColor(100.0f/255.0f, 149.0f/255.0f, 237.0f/255.0f, 1.0f); // Cornflower Blue
	GL.glClear(GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT);
	GL.glLoadIdentity();

	camera.lookAt();
	
	for (int y = -1; y <= theWorld.getHeight(); ++y)
	{
	    for (int x = -1; x <= theWorld.getWidth(); ++x)
	    {
		if (theWorld.isWall(new GridPoint(x,y)))
		    this.drawWall(x, y, 1.0f, 1.0f);
		
		this.drawFloor(x, y, 1.0f, 1.0f);
	    }
	}

	GL.glDisable(GL.GL_TEXTURE_2D);

	this.drawEnemy(enemy.x, enemy.y, 1.0f, 1.0f);
	this.drawGoal(goal.x, goal.y, 1.0f, 1.0f);

	GL.glDisable(GL.GL_DEPTH_TEST);
	GL.glFlush();
    }

    protected void drawEnemy(float x, float z, float width, float depth)
    {
	final float kBottom = -0.15f;
	final float kTop = 0.15f;

	GL.glColor3f(0.68f, 0.12f, 0.16f);
	GL.glBegin(GL.GL_QUADS);

	x += 0.3f;
	z += 0.3f;
	width = 0.4f;
	depth = 0.4f;

	// Front
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);

	// Back
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x,          kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width,  kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width,  kTop, z + depth);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x,          kTop, z + depth);

	// Left
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x, kTop, z + depth);

	// Right
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z + depth);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);

	// Top
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x + width, kTop, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kTop, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x, kTop, z + depth);

	GL.glEnd();
    }

    protected void drawGoal(float x, float z, float width, float depth)
    {
	final float kBottom = -0.5f;
	final float kTop = -0.25f;

	x += 0.4f;
	z += 0.4f;
	
	width = 0.2f;
	depth = 0.2f;

	GL.glColor3f(1.0f, 1.0f, 0.0f);
	GL.glBegin(GL.GL_QUADS);

	// Front
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);

	// Back
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x,          kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width,  kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width,  kTop, z + depth);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x,          kTop, z + depth);

	// Left
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x, kTop, z + depth);

	// Right
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z + depth);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);

	GL.glEnd();
    }

    protected void drawWall(float x, float z, float width, float depth)
    {
	final float kBottom = -0.5f;
	final float kTop = 0.35f;

	// Draw the Cell

	if(theTextures != null)
        {
	    if (theTextures[0] != null)
		theTextures[0].bind();
       	}

	GL.glColor3f(1.0f, 1.0f, 1.0f);
	GL.glBegin(GL.GL_QUADS);

	// Front
	GL.glNormal3f(0.0f,0.0f,-1.0f);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);

	// Back
	GL.glNormal3f(0.0f,0.0f,1.0f);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x,          kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width,  kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width,  kTop, z + depth);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x,          kTop, z + depth);

	// Left
	GL.glNormal3f(-1.0f,0.0f,0.0f);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kTop, z);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x, kTop, z + depth);

	// Right
	GL.glNormal3f(1.0f,0.0f,0.0f);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z + depth);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z + depth);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kTop, z);

	GL.glEnd();
    }

    public void drawLabel(float x, float y, float width, float height, String theString)
    {
	Rectangle2D theBounds = theTextRenderer.getBounds(theString);
	theTextRenderer.begin3DRendering();
	theTextRenderer.setColor(0.0f, 1.0f, 0.0f, 1.0f);
	theTextRenderer.draw3D(theString, 
			       x + width/2.0f - (float)theBounds.getCenterX(), 
			       0.0f,
			       y + height/2.0f - (float)theBounds.getCenterY(), 
			       0.25f);
	theTextRenderer.end3DRendering();
    }

    protected void drawFloor(float x, float z, float width, float depth)
    {
	if(theTextures != null)
        {
	    if (theTextures[1] != null)
		theTextures[1].bind();
       	}

	final float kBottom = -0.5f;
	final float xStart = 4.0f/10.0f;
	final float xFinish = 6.0f/10.0f;

	final float yStart = 4.0f/10.0f;
	final float yFinish = 6.0f/10.0f;

	GL.glColor3f(0.5f, 0.5f, 0.5f);
	GL.glBegin(GL.GL_LINES);
	GL.glVertex3f(x + (xStart * width),  kBottom + 0.05f, z + (depth / 2.0f)); 
	GL.glVertex3f(x + (xFinish * width), kBottom + 0.05f, z + (depth / 2.0f)); 
	GL.glVertex3f(x + (width/2.0f),      kBottom + 0.05f, z + (yStart * depth)); 
	GL.glVertex3f(x + (width/2.0f),      kBottom + 0.05f, z + (yFinish * depth)); 
	GL.glEnd();

	// Floor
	GL.glColor3f(1.0f, 1.0f, 1.0f);
	GL.glBegin(GL.GL_QUADS);
	GL.glTexCoord2f(0.0f, 0.0f);
	GL.glVertex3f(x, kBottom, z);
	GL.glTexCoord2f(1.0f, 0.0f);
	GL.glVertex3f(x + width, kBottom, z);
	GL.glTexCoord2f(1.0f, 1.0f);
	GL.glVertex3f(x + width, kBottom, z + depth);
	GL.glTexCoord2f(0.0f, 1.0f);
	GL.glVertex3f(x, kBottom, z + depth);
	GL.glEnd();
    }
}
