import java.util.Random;
import java.util.Vector;
import java.util.Collections;
import java.util.ArrayList;

public class GridWorld implements Environment
{
    public static int kEmpty = 0;
    public static int kWall  = 2;
    public static int kGoal  = 4;
    public static int kAgent = 8;
    public static int kEnemy = 16;

    public static final double caughtReward = -100;
    public static final double goalReward = 100;
    public static final double stepReward = 1.0;
    
    public static int kUp        = 0;
    public static int kDown      = 1;
    public static int kLeft      = 2;
    public static int kRight     = 3;
    public static int kNumActions= 4;

    protected Agent theAgent;
    protected int[] theWorld = null;
    protected double[] theValues = null;
    protected int theWidth, theHeight;

    protected GridPoint theStartPosition;
    protected GridPoint theGoalPosition;
    protected GridPoint theAgentPosition;
    protected GridPoint theEnemyPosition;

    protected Random theGenerator;

    protected int episodeNum;
    protected double score;

		protected double[] last_hundred_scores = new double[100];
		protected int last_hundred_scores_next_slot = 0;


    protected double getDistance(GridPoint a, GridPoint b)
    {
				int dx = (a.x - b.x);
				int dy = (a.y - b.y);
	
				return dx*dx + dy*dy;
    }

    protected int getIndexFromPoint(GridPoint thePoint)
    {
				return thePoint.y * this.getWidth() + thePoint.x;
    }

    public GridWorld(int theWidth, int theHeight)
    {
				this.theWidth = theWidth;
				this.theHeight = theHeight;
				this.theWorld = new int[this.theWidth*this.theHeight];
				this.theValues = new double[this.theWidth*this.theHeight];

				this.theStartPosition = new GridPoint();
				this.theGoalPosition  = new GridPoint();
				this.theAgentPosition = new GridPoint();
				this.theEnemyPosition = new GridPoint();
    
				this.theGenerator = new Random();

				this.episodeNum = 0;
				this.score = 0;
    }

    public void setAgent(Agent theAgent)
    {
				this.theAgent = theAgent;
    }

    public Agent getAgent()
    {
				return this.theAgent;
    }

    public void removeWall(GridPoint thePoint)
    {
				final int i = this.getIndexFromPoint(thePoint);
				this.theWorld[i] &= ~GridWorld.kWall;
    }

    public void addWall(GridPoint thePoint)
    {
				final int i = this.getIndexFromPoint(thePoint);
				this.theWorld[i] |= GridWorld.kWall;
    }

    public void setStartPosition(GridPoint thePoint)
    {
				this.theStartPosition.x = thePoint.x;
				this.theStartPosition.y = thePoint.y;
    }

    public void setGoalPosition(GridPoint thePoint)
    {
				this.theGoalPosition.x = thePoint.x;
				this.theGoalPosition.y = thePoint.y;
    }

    public void setAgentPosition(GridPoint thePoint)
    {
				//System.err.println("Setting agent position to "+thePoint.x+","+thePoint.y);
				this.theAgentPosition.x = thePoint.x;
				this.theAgentPosition.y = thePoint.y;
				//System.err.println("Set agent position to "+theAgentPosition.x+","+theAgentPosition.y);
    }

    public void setEnemyPosition(GridPoint thePoint)
    {
				this.theEnemyPosition.x = thePoint.x;
				this.theEnemyPosition.y = thePoint.y;
    }

    public GridPoint getStartPosition()
    {
				return this.theStartPosition;
    }

    public GridPoint getGoalPosition()
    {
				return this.theGoalPosition;
    }

    public GridPoint getAgentPosition()
    {
				//System.err.println(" Get agent position "+this.theAgentPosition.x+", "+this.theAgentPosition.y);
				return this.theAgentPosition;
    }

    public GridPoint getEnemyPosition()
    {
				return this.theEnemyPosition;
    }

    public boolean isWall(GridPoint thePoint)
    {
				final int i = this.getIndexFromPoint(thePoint);
				boolean isWall = false;

				if (thePoint.x < 0 || thePoint.x >= this.getWidth() ||
						thePoint.y < 0 || thePoint.y >= this.getHeight())

						isWall = true;
	
				else if ((this.theWorld[i] & GridWorld.kWall) != 0)
						isWall = true;
	
				return isWall;
    }

    public boolean isStartPosition(GridPoint thePoint)
    {
				return this.theStartPosition.equals(thePoint);
    }

    public boolean isGoalPosition(GridPoint thePoint)
    {
				return this.theGoalPosition.equals(thePoint);
    }

    public boolean isAgentPosition(GridPoint thePoint)
    {
				return this.theAgentPosition.equals(thePoint);
    }

    public boolean isEnemyPosition(GridPoint thePoint)
    {
				return this.theEnemyPosition.equals(thePoint);
    }

    public int getWidth()
    {
				return this.theWidth;
    }

    public int getHeight()
    {
				return this.theHeight;
    }

    public Integer getStateNumber(int ax, int ay, int ex, int ey)
    {
				final int C = this.getWidth();
				final int B = this.getHeight() * C;
				final int A = this.getWidth()  * B;

				int theState = 
						ay * A + 
						ax * B +
						ey * C +
						ex;

				return new Integer(theState);
    }

    public void updateValues()
    {
				for (int y = 0; y < this.getHeight(); ++y)
						{
								for (int x = 0; x < this.getWidth(); ++x)
										{
												GridPoint ep = this.getEnemyPosition();
												int theState = this.getStateNumber(x,y,ep.x,ep.y).intValue();

												int theAction = 0;
												QBasedAgent theAgent = (QBasedAgent)this.theAgent;
												for(int a = 1; a < this.kNumActions; ++a)
														{
																if (theAgent.Q(theState,a) > theAgent.Q(theState,theAction))
																		theAction = a;
														}

												this.theValues[y*this.getWidth()+x] = theAgent.Q(theState,theAction);
										}
						}
    }

    public double getValue(GridPoint p)
    {
				return this.theValues[this.getWidth() * p.y + p.x];
    }

    public double getScore()
    {
				return this.score;
    }

    public int getEpisodeNum()
    {
				return this.episodeNum;
    }

    public void incrementEpisode()
    {
				//	System.err.println("Incrementing episode");
				this.episodeNum++;
				this.score = 0;
    }

    public void addScore(double d)
    {
				//System.err.println("Adding "+d+" to score");
				this.score+=d;
    }

    /* Environment Interface */
    public void init()
    {
    }

    public Object start()
    {
				//this.theAgentPosition = new GridPoint(this.theStartPosition);
				return this.getStateNumber(this.theAgentPosition.x, this.theAgentPosition.y,
																	 this.theEnemyPosition.x, this.theEnemyPosition.y);
    }

    public void moveEnemy()
    {
				/* Move the enemy */
				GridPoint newEnemyPosition      = new GridPoint(this.theEnemyPosition);
				GridPoint newEnemyPositionLeft  = new GridPoint(this.theEnemyPosition.x - 1, this.theEnemyPosition.y);
				GridPoint newEnemyPositionRight = new GridPoint(this.theEnemyPosition.x + 1, this.theEnemyPosition.y);
				GridPoint newEnemyPositionDown  = new GridPoint(this.theEnemyPosition.x,     this.theEnemyPosition.y - 1);
				GridPoint newEnemyPositionUp    = new GridPoint(this.theEnemyPosition.x,     this.theEnemyPosition.y + 1);

				double currentDistance = this.getDistance(this.theAgentPosition, newEnemyPosition);
				double leftDistance    = this.getDistance(this.theAgentPosition, newEnemyPositionLeft);
				double rightDistance   = this.getDistance(this.theAgentPosition, newEnemyPositionRight);
				double downDistance    = this.getDistance(this.theAgentPosition, newEnemyPositionDown);
				double upDistance      = this.getDistance(this.theAgentPosition, newEnemyPositionUp);
	
				//See if there's a move available that narrows the distance.
				ArrayList positions = new ArrayList();
				if(leftDistance<=currentDistance && !isWall(newEnemyPositionLeft))
						positions.add(newEnemyPositionLeft);
				if(rightDistance<=currentDistance && !isWall(newEnemyPositionRight))
						positions.add(newEnemyPositionRight);
				if(upDistance<=currentDistance && !isWall(newEnemyPositionUp)) 
						positions.add(newEnemyPositionUp);
				if(downDistance<=currentDistance && !isWall(newEnemyPositionDown)) 
						positions.add(newEnemyPositionDown);

				//make any move that narrows the distance
				if(!positions.isEmpty()) {
						Collections.shuffle(positions);
						newEnemyPosition = (GridPoint)positions.get(0);
				}
				//if not available, make a random move
				else {
						int randAction = theGenerator.nextInt(4);
						if (randAction == 0)
								newEnemyPosition = newEnemyPositionLeft;
						else if (randAction == 1)
								newEnemyPosition = newEnemyPositionRight;
						else if (randAction == 2)
								newEnemyPosition = newEnemyPositionDown;
						else
								newEnemyPosition = newEnemyPositionUp;
				}
				/*
          
        int bestIndex;
        double bestValue, value;
        int numBests = 1;   

        bestIndex = 0;
        bestValue = Q[s][0];	
    
        for(int i = 1; i < this.numActions; ++i)
        {
				value = Q[s][i];
				if( value >= bestValue )
				{
				bestValue = value;
				bestIndex = i;
				numBests = 1;
				} 
        }
        return bestIndex;        


				System.err.println("Cur: "+currentDistance+" Left"+leftDistance+" Right"+rightDistance+" Down:"+downDistance+" Up"+upDistance);
				/*
				if (leftDistance < rightDistance && 
				leftDistance < downDistance  && 
				leftDistance < upDistance &&
				leftDistance < currentDistance && 
				!isWall(newEnemyPositionLeft))
				newEnemyPosition = newEnemyPositionLeft;

				else if (rightDistance < leftDistance   && 
				rightDistance < downDistance  && 
				rightDistance < upDistance &&
				rightDistance < currentDistance &&
				!isWall(newEnemyPositionRight))
				newEnemyPosition = newEnemyPositionRight;

				else if (downDistance < leftDistance   && 
				downDistance < rightDistance  && 
				downDistance < upDistance &&
				downDistance < currentDistance &&
				!isWall(newEnemyPositionDown))
				newEnemyPosition = newEnemyPositionDown;

				else if (upDistance < leftDistance && 
				upDistance < rightDistance  && 
				upDistance < downDistance &&
				upDistance < currentDistance &&
				!isWall(newEnemyPositionUp))
				newEnemyPosition = newEnemyPositionUp;

				else
				{
				System.err.println("Picking random direction");
				int randAction = theGenerator.nextInt(4);
				if (randAction == 0)
				newEnemyPosition = newEnemyPositionLeft;
				else if (randAction == 1)
				newEnemyPosition = newEnemyPositionRight;
				else if (randAction == 2)
				newEnemyPosition = newEnemyPositionDown;
				else
				newEnemyPosition = newEnemyPositionUp;
				}

				*/
				if (!this.isWall(newEnemyPosition))
						this.theEnemyPosition = newEnemyPosition;
    }
		
    public void moveAgent(int action)
    {

				/* Make a backup of the agents position */
				GridPoint newAgentPosition = new GridPoint(this.theAgentPosition);

				/* Let the agent make its move */
				if (action == GridWorld.kLeft)
						newAgentPosition.x -= 1;
	
				else if (action == GridWorld.kRight)
						newAgentPosition.x += 1;
	
				else if (action == GridWorld.kUp)
						newAgentPosition.y -= 1;
	
				else if (action == GridWorld.kDown)
						newAgentPosition.y += 1;

				/* Now check if the agent's new position is in a wall, and if so, back him up. */
				if (!this.isWall(newAgentPosition))
						this.theAgentPosition = newAgentPosition;
    }

    public boolean isCaptured()
    {
				return
						this.theAgentPosition.x == this.theEnemyPosition.x &&
						this.theAgentPosition.y == this.theEnemyPosition.y;
    }

    public boolean isAtGoal()
    {
				return
						this.theAgentPosition.x == this.theGoalPosition.x &&
						this.theAgentPosition.y == this.theGoalPosition.y;
    }

    public void reset()
    {
				Random openSpaceGenerator = new Random();
				boolean reset = false;

				Vector openSpace = new Vector();
	
				for(int y = 0; y < this.getHeight(); ++y)
						{
								for(int x = 0; x < this.getWidth(); ++x)
										{
												GridPoint p = new GridPoint(x,y);
												if (!this.isWall(p))
														openSpace.addElement(p);
										}
						}

				int agentStart = openSpaceGenerator.nextInt(openSpace.size());
				int enemyStart = openSpaceGenerator.nextInt(openSpace.size());
				//int goalStart = openSpaceGenerator.nextInt(openSpace.size());

				if (enemyStart == agentStart && openSpace.size() >= 2)
						{
								while (enemyStart == agentStart)
										enemyStart = openSpaceGenerator.nextInt(openSpace.size());
						}

				this.setAgentPosition((GridPoint)openSpace.elementAt(agentStart));
				this.setEnemyPosition((GridPoint)openSpace.elementAt(enemyStart));
				//this.setGoalPosition((GridPoint)openSpace.elementAt(goalStart));
				//System.err.println("Agent start: "+this.theAgentPosition.x+", "+this.theAgentPosition.y);
				last_hundred_scores[last_hundred_scores_next_slot] = score;
				last_hundred_scores_next_slot = (last_hundred_scores_next_slot + 1) % last_hundred_scores.length;
				double window_avg = 0;
				double window_max = -100;
				for(int i = 0; i < last_hundred_scores.length; i++) {
						window_max = Math.max(last_hundred_scores[i], window_max);
						window_avg += last_hundred_scores[i];
				}
				window_avg /= last_hundred_scores.length;

				System.out.println(episodeNum+"\t"+score+"\t"+window_avg+"\t"+window_max);

				this.incrementEpisode();
		}

    public Object[] step(Integer Action)
    {
				boolean captured = false;
				Object[] ORT = new Object[3];
				int action = Action.intValue();
				Integer theStateNumber = this.getStateNumber(this.theAgentPosition.x, this.theAgentPosition.y,
																										 this.theEnemyPosition.x, this.theEnemyPosition.y);

	
				this.moveAgent(action);
				captured = isCaptured();

				this.moveEnemy();
				captured = captured || isCaptured();

				this.updateValues();

				Double theReward;

				if (captured == true)
						{
								theReward = new Double(caughtReward);
								addScore(caughtReward);
						}
				else if (isAtGoal() == true)
						{
								theReward = new Double(goalReward);
								addScore(goalReward);
						}
				else
						{
								theReward = new Double(stepReward);
								addScore(stepReward);
						}

				ORT[0] = theStateNumber;
				ORT[1] = theReward;
				ORT[2] = new Boolean(captured);

				return ORT;
    }

    public Double end()
    {
				return new Double(0.0);
    }
}
