import java.util.*;
/*
 *  QlearningAgent.java
 *  
 *
 *  Created by Adam White onNov 21 2005.
 *  Copyright (c) 2005 UAlberta. All rights reserved.
 * 
 *  A GENERAL simple tabluar agent that uses sarsa policy iteration.
 *  Desgined for Episodic tasks. 
 *  
 *  State: Integer Single Dimension
 *  Actions: Integer Single dimension
 *
 */ 
 
public class QlearningAgent implements Agent, QBasedAgent
{
    private int A, S;
    private int [] actions;
    private double [][] Q;
    private double alpha = .5, epsilon = .01;
    private int numActions;
    private int numStates;
    private double gamma = 0.9;

    QlearningAgent(int numActions, int numStates)
    {
				this.numActions = numActions;
				this.numStates = numStates;     
    }

    public void setAlpha(double a) 
    {
				alpha = a;
    }

    public void setEpsilon(double e)
    {
				epsilon = e;
    }

    public double getAlpha()
    {
				return alpha;
    }

    public double getEpsilon()
    {
				return epsilon;
    }

    public void resetValue()
    {
        for( int i = 0; i < this.numStates; i++)
            for(int j =0; j < this.numActions; j++)
                Q[i][j] = 1; 
                
    }

    public void init()
    {
        actions =  new int[this.numActions];
        Q = new double[this.numStates][this.numActions];
        
        for (int i = 0; i < this.numActions; ++i)
            actions[i] = i;
            
				resetValue();
    }            
    
    public Integer start(Integer obs)
    {
        S = obs.intValue();
        A = egreedy(S);
        
        return new Integer(actions[A]);     
    }

    
    public Integer step(Double R, Integer obs)
    {
				double r;
        double delta;
        int iNew, iOld, a, s;
            
				r = R.doubleValue();
        s = obs.intValue();
        int amax = argmax(s);
				a = egreedy(s);

        delta = r + gamma*Q[s][amax] - Q[S][A];
        Q[S][A] = Q[S][A] + alpha*delta;
        A = a;
        S = s;
    
        return new Integer(actions[a]);
    }
    
    public void end(Double R)
    {
        int i;
        double error;
				double r = R.doubleValue();

        error = r - Q[S][A];
        Q[S][A] = Q[S][A] + alpha*error;
    }    
    
    public void cleanup()
    {
    }
    
    public int egreedy(int s)
    {
        double ran = Math.random();
        if( ran > epsilon ) 
						{
								return argmax(s);
						}
        else
						{
								return (int)(Math.random()*this.numActions);
						}
    }
    
    public int argmax(int s)
    {
        //find the index of the maximal action, break ties randomly
        int bestIndex;
        double bestValue, value;
        int numBests = 1;   
          
        bestIndex = 0;
        bestValue = Q[s][0];	
    
        for(int i = 1; i < this.numActions; ++i)
						{
                value = Q[s][i];
                if( value >= bestValue )
										{
												bestValue = value;
												bestIndex = i;
												numBests = 1;
										} 
						}
        return bestIndex;        
    }

    public double Q(int s, int a)
    {
				return Q[s][a];
    }
}
